import numpy as np
import networkx as nx
from community import community_louvain
import matplotlib.pyplot as plt
from scipy.optimize import bisect

class Networks(object):
    """docstring for Networks."""

# ------------------------------- INITIALIZER -------------------------------- #
    def __init__(self, cells, filtered_slow, filtered_fast, ND_avg=8):
        self.__ND_avg = ND_avg
        self.__cells = cells
        self.__filtered_slow = filtered_slow
        self.__filtered_fast = filtered_fast

        self.__R_slow = False
        self.__R_fast = False

        self.__A_slow = False
        self.__A_fast = False

        self.__G_slow = False
        self.__G_fast = False

# --------------------------------- GETTERS ---------------------------------- #

    def get_G_slow(self): return self.__G_slow
    def get_G_fast(self): return self.__G_fast
    def get_R_slow(self): return self.__R_slow
    def get_R_fast(self): return self.__R_fast
    def get_A_slow(self): return self.__A_slow
    def get_A_fast(self): return self.__A_fast

# ----------------------------- NETWORK METHODS ------------------------------ #
    def build_networks(self):
        # Compute correlation matrices
        self.__construct_correlation_matrix()
        # Calculate threshold and construct network
        slow_threshold, r = bisect(lambda x: self.__graph_from_threshold(self.__R_slow, x)["ND"]-self.__ND_avg, 0, 1, full_output=True)
        # Calculate threshold and construct network
        fast_threshold, r = bisect(lambda x: self.__graph_from_threshold(self.__R_fast, x)["ND"]-self.__ND_avg, 0, 1, full_output=True)

        # self.__G_slow = self.__graph_from_threshold(self.__R_slow, slow_threshold)["G"]
        self.__G_slow = self.__graph_from_threshold(self.__R_slow, slow_threshold)["G"]
        self.__G_fast = self.__graph_from_threshold(self.__R_fast, fast_threshold)["G"]

        self.__A_slow = nx.to_numpy_matrix(self.__G_slow)
        self.__A_fast = nx.to_numpy_matrix(self.__G_fast)

    def __construct_correlation_matrix(self):
        self.__R_slow, self.__R_fast = np.eye(self.__cells), np.eye(self.__cells)
        for i in range(self.__cells):
            for j in range(i):
                corr_slow = np.corrcoef(self.__filtered_slow[i], self.__filtered_slow[j])[0,1]
                self.__R_slow[i,j], self.__R_slow[j,i] = corr_slow, corr_slow
                corr_fast = np.corrcoef(self.__filtered_fast[i], self.__filtered_fast[j])[0,1]
                self.__R_fast[i,j], self.__R_fast[j,i] = corr_fast, corr_fast

    def __graph_from_threshold(self, R, R_threshold):
        G = nx.Graph()
        for i in range(self.__cells):
            G.add_node(i)
            for j in range(i):
                if R[i,j] >= R_threshold:
                    G.add_edge(i,j)
        ND = np.mean([G.degree(i) for i in G])
        return {"G": G, "ND": ND}

    def node_degree(self, cell):
        return (self.__G_slow.degree(cell),
                self.__G_fast.degree(cell))

    def clustering(self, cell):
        return (nx.clustering(self.__G_slow)[cell],
                nx.clustering(self.__G_fast)[cell])

    def nearest_neighbour_degree(self, cell):
        return (nx.average_neighbor_degree(self.__G_slow)[cell],
                nx.average_neighbor_degree(self.__G_fast)[cell])

    def modularity(self):
        partition_slow = community_louvain.best_partition(self.__G_slow)
        partition_fast = community_louvain.best_partition(self.__G_fast)
        Q_slow = community_louvain.modularity(partition_slow, self.__G_slow)
        Q_fast = community_louvain.modularity(partition_fast, self.__G_fast)

        return (Q_slow, Q_fast)

    def global_efficiency(self):
        GE_slow = nx.global_efficiency(self.__G_slow)
        GE_fast = nx.global_efficiency(self.__G_fast)
        return (GE_slow, GE_fast)

    def max_connected_component(self):
        MS_slow = len(max(nx.connected_components(self.__G_slow), key=len))/self.__cells
        MS_fast = len(max(nx.connected_components(self.__G_fast), key=len))/self.__cells
        return (MS_slow, MS_fast)

    def average_correlation(self):
        R_slow = np.matrix(self.__R_slow)
        R_fast = np.matrix(self.__R_fast)
        R_slow_upper = R_slow[np.triu_indices(R_slow.shape[0])]
        R_fast_upper = R_fast[np.triu_indices(R_fast.shape[0])]
        return (R_slow_upper.mean(), R_fast_upper.mean())

    def draw_networks(self, positions, ax1, ax2, colors):
        nx.draw(self.__G_slow, pos=positions, ax=ax1, with_labels=False, node_size=50, width=0.25, font_size=3, node_color=colors[0])
        nx.draw(self.__G_fast, pos=positions, ax=ax2, with_labels=False, node_size=50, width=0.25, font_size=3, node_color=colors[1])



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import matplotlib.gridspec as gridspec
import matplotlib.patches as patches
import matplotlib.colors as mc
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter, AutoMinorLocator)

import pickle
from pathlib import Path
import colorsys

from langerhans.analysis import Analysis


def lighten_color(color, amount=0.5):
    """
    Lightens the given color by multiplying (1-luminosity) by the given amount.
    Input can be matplotlib color string, hex string, or RGB tuple.

    Examples:
    >> lighten_color('g', 0.3)
    >> lighten_color('#F034A3', 0.6)
    >> lighten_color((.3,.55,.1), 0.5)
    """
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])

LOW_LIST = ["white", "C0", lighten_color("C0", 1.5)]
HIGH_LIST = ["white", "C3", lighten_color("C3", 1.5)]
LOW_CMAP = mc.LinearSegmentedColormap.from_list("", LOW_LIST)
HIGH_CMAP = mc.LinearSegmentedColormap.from_list("", HIGH_LIST)
DARK_BLUE = lighten_color("C0", 1.3)
LIGHT_BLUE = lighten_color("C0", 0.8)
DARK_RED = lighten_color("C3", 1.3)
LIGHT_RED = lighten_color("C3", 0.8)

class GlobalAnalysis(object):
    """docstring for GlobalAnalysis."""

    def __init__(self, path):
        data_path = Path(path)
        if not data_path.exists() or not data_path.is_dir():
            raise ValueError("Directory does not exist.")

        pkl_file_list = [x for x in data_path.glob("*.pkl")]
        pickle_file_list = [x for x in data_path.glob("*.pickle")]
        data_list = pkl_file_list + pickle_file_list
        if len(data_list) == 0:
            raise ValueError("No pickle files found.")

        positions_list = []
        for s in data_list:
            positions = data_path / Path("{0}.txt".format(s.stem))
            if not positions.exists():
                raise ValueError("Positions file for {0} not found.".format(s.stem))
            positions_list.append(positions)

        self.__data_dict = {}
        self.__low_glucose = []
        self.__high_glucose = []
        self.__pars_network = {}
        self.__pars_cell = {}
        self.__spikes_v_phases = {}
        self.__spikes_v_phases_sep = {}
        self.__corr_v_dist = {}
        self.__networks = {}

        for d, p in zip(data_list, positions_list):
            s = d.stem
            self.__data_dict[s] = d
            with d.open("rb") as f:
                data = pickle.load(f)
            with p.open() as f:
                positions = np.loadtxt(f)

            glc = data.get_settings()["Glucose [mM]"]
            if glc == 8:
                self.__low_glucose.append(s)
            elif glc == 12:
                self.__high_glucose.append(s)

            analysis = Analysis()
            analysis.import_data(data, positions)
            analysis.build_networks()
            network = analysis.get_networks()

            self.__pars_cell[s], self.__pars_network[s] = analysis.compute_parameters()
            self.__spikes_v_phases[s] = analysis.spikes_vs_phase()
            self.__spikes_v_phases_sep[s] = analysis.spikes_vs_phase(mode="separate")
            self.__corr_v_dist[s] = analysis.correlation_vs_distance()
            self.__networks[s] = (network.get_G_slow(), network.get_G_fast())

            del data
            del analysis

    def get_pars_network(self): return self.__pars_network
    def get_pars_cell(self): return self.__pars_cell

    def get_data(self, series):
        path = self.__data_dict[series]
        with path.open("rb") as f:
            data = pickle.load(f)
        return data

    def mean_std_local(self, parameter):
        means_low = [np.nanmean([self.__pars_cell[s][c][parameter] for c in range(len(self.__pars_cell[s]))]) for s in self.__low_glucose]
        means_high = [np.nanmean([self.__pars_cell[s][c][parameter] for c in range(len(self.__pars_cell[s]))]) for s in self.__high_glucose]

        mean_low = np.mean(means_low)
        mean_high = np.mean(means_high)
        std_low = np.std(means_low)
        std_high = np.std(means_high)

        return ((means_low, means_high),(mean_low, mean_high),(std_low, std_high))

    def mean_std_global(self, parameter):
        values_low = [self.__pars_network[s][parameter] for s in self.__low_glucose]
        values_high = [self.__pars_network[s][parameter] for s in self.__high_glucose]
        mean_low = np.mean(values_low)
        mean_high = np.mean(values_high)
        std_low = np.std(values_low)
        std_high = np.std(values_high)

        return ((values_low, values_high), (mean_low, mean_high), (std_low, std_high))

    def plot_avg_stds_local(self, ax, parameter):
        values, means, stds = self.mean_std_local(parameter)

        g_low = [1 for i in range(len(values[0]))]
        g_high = [2 for i in range(len(values[1]))]

        ax.bar([1,2], means, 0.5, yerr=stds, color=(DARK_BLUE, DARK_RED))
        ax.scatter(g_low+g_high, values[0]+values[1], c="k", zorder=10, s=5)
        ax.set_xticks([1,2])
        ax.set_xticklabels(("8 mM", "12 mM"))
        ax.set_xlabel("Glucose concentration (mM)")
        ax.set_ylabel(parameter)

    def plot_avg_stds_global(self, ax, parameter, glucose=False):
        if glucose is not False:
            values0, means0, stds0 = self.mean_std_global(parameter[0])
            values1, means1, stds1 = self.mean_std_global(parameter[1])
            if glucose == 8:
                values = (values0[0], values1[0])
                means = (means0[0], means1[0])
                stds = (stds0[0], stds1[0])
                color = (LIGHT_BLUE, DARK_BLUE)
            elif glucose == 12:
                values = (values0[1], values1[1])
                means = (means0[1], means1[1])
                stds = (stds0[1], stds1[1])
                color = (LIGHT_RED, DARK_RED)
            ax.set_xticklabels(parameter)
            ax.set_xlabel("Parameter")

        else:
            values, means, stds = self.mean_std_global(parameter) # Low/high glucose values
            color = (DARK_BLUE, DARK_RED)
            ax.set_xticklabels(("8 mM", "12 mM"))
            ax.set_xlabel("Glucose concentration (mM)")

        x_left = [1 for i in range(len(values[0]))]
        x_right = [2 for i in range(len(values[1]))]

        ax.bar([1,2], means, 0.5, yerr=stds, color=color)
        ax.scatter(x_left+x_right, values[0]+values[1], c="k", zorder=10, s=5)
        ax.set_xticks([1,2])

    def plot_spikes_vs_phases(self, ax, series):
        if series == "low":
            phases = self.__spikes_v_phases[self.__low_glucose[0]][0]
            spikes = [self.__spikes_v_phases[s][1] for s in self.__low_glucose]
            cmap = LOW_CMAP
            spikes = np.sum(spikes, axis=0)
        elif series == "high":
            phases = self.__spikes_v_phases[self.__high_glucose[0]][0]
            spikes = [self.__spikes_v_phases[s][1] for s in self.__high_glucose]
            cmap = HIGH_CMAP
            spikes = np.sum(spikes, axis=0)
        elif series == "both":
            phases = self.__spikes_v_phases[self.__low_glucose[0]][0]
            spikes = [self.__spikes_v_phases[s][1] for s in self.__low_glucose+self.__high_glucose]
            cmap = HIGH_CMAP
            spikes = np.sum(spikes, axis=0)
        elif series in self.__low_glucose or series in self.__high_glucose:
            phases, spikes = self.__spikes_v_phases[series]
            cmap = LOW_CMAP if series in self.__low_glucose else HIGH_CMAP
        norm_spikes = spikes/np.max(spikes)
        colors = cmap(norm_spikes)
        ax.bar(phases, norm_spikes, width=2*np.pi/12, bottom=0.0, color=colors)
        ax.set_thetagrids(angles=range(0,360,30))

    def plot_corr_vs_dist(self, ax, bin_number=15, max_distance=210, mode="both"):
        distances_left = []
        distances_right = []
        correlations_left = []
        correlations_right = []
        if mode in ("low", "high", "both"):
            labels = ("slow", "fast")
            if mode == "low":
                series = self.__low_glucose
                color = (LIGHT_BLUE, DARK_BLUE)
            elif mode == "high":
                series = self.__high_glucose
                color = (LIGHT_RED, DARK_RED)
            elif mode == "both":
                series = self.__low_glucose + self.__high_glucose
                color = (LIGHT_RED, DARK_RED)
            for s in series:
                distances_left.extend(self.__corr_v_dist[s][0])
                distances_right.extend(self.__corr_v_dist[s][0])
                correlations_left.extend(self.__corr_v_dist[s][1])
                correlations_right.extend(self.__corr_v_dist[s][2])
        elif mode in ("slow", "fast"):
            labels = ("low", "high")
            if mode == "slow":
                color = (LIGHT_BLUE, LIGHT_RED)
                for s in self.__low_glucose:
                    distances_left.extend(self.__corr_v_dist[s][0])
                    correlations_left.extend(self.__corr_v_dist[s][1])
                for s in self.__high_glucose:
                    distances_right.extend(self.__corr_v_dist[s][0])
                    correlations_right.extend(self.__corr_v_dist[s][1])
            elif mode == "fast":
                color = (DARK_BLUE, DARK_RED)
                for s in self.__low_glucose:
                    distances_left.extend(self.__corr_v_dist[s][0])
                    correlations_left.extend(self.__corr_v_dist[s][2])
                for s in self.__high_glucose:
                    distances_right.extend(self.__corr_v_dist[s][0])
                    correlations_right.extend(self.__corr_v_dist[s][2])


        bins_left = np.linspace(0, max_distance, bin_number)
        inds_left = np.digitize(distances_left, bins_left)

        bins_right = np.linspace(0, max_distance, bin_number)
        inds_right = np.digitize(distances_right, bins_right)

        left_mean = np.mean(correlations_left)
        right_mean = np.mean(correlations_right)

        left_binarized = [[] for b in range(bin_number)]
        right_binarized = [[] for b in range(bin_number)]


        for i in range(len(inds_left)):
            left_binarized[inds_left[i]-1].append(correlations_left[i])
        for i in range(len(inds_right)):
            right_binarized[inds_right[i]-1].append(correlations_right[i])

        left = []
        right = []
        for i in range(bin_number):
            left.append(np.mean(left_binarized[i]))
            right.append(np.mean(right_binarized[i]))

        x = np.arange(bin_number)
        width = 0.35  # the width of the bars

        rects1 = ax.bar(x[:-1] - 0.5 - width/2, left[:-1], width, color=color[0], label=labels[0])
        rects2 = ax.bar(x[:-1] - 0.5 + width/2, right[:-1], width, color=color[1], label=labels[1])

        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax.set_ylabel('Average correlation coefficient')
        ax.set_xlabel("Distance [μm]")
        ax.set_xticks(x)
        ax.set_xticklabels(np.around(bins_left[1:],0).astype(int), rotation=45)
        ax.set_ylim(0.2,0.9)
        ax.yaxis.set_major_locator(MultipleLocator(0.2))

        # For the minor ticks, use no labels; default NullFormatter.
        ax.yaxis.set_minor_locator(MultipleLocator(0.1))
        ax.legend()



import numpy as np

from scipy.signal import butter, sosfiltfilt
from scipy.stats import skew, skewnorm, norm
from scipy.optimize import curve_fit, differential_evolution

import matplotlib.pyplot as plt
from matplotlib.colors import colorConverter
import matplotlib.transforms as transforms
import matplotlib.patches as patches

EXCLUDE_COLOR = 'xkcd:salmon'
SAMPLE_SETTINGS = {
    "Glucose [mM]": 8,
    "Sampling [Hz]": 10,
    "Stimulation [frame]": [1200, 0],
    "Filter":
        {
        "Slow [Hz]": [0.001, 0.005],
        "Fast [Hz]": [0.04, 0.4],
        "Plot [s]": [250, 1750]
        },
    "Distribution order": 5,
    "Exclude":
        {
        "Score threshold": 1,
        "Spikes threshold": 0.01
        },
    "Distance [um]": 1,
    "Plateau [s]": [0, 0]
    }
STD_RATIO = 2


class Data(object):
    """
    A class for signal analysis.
    """
# ------------------------------- INITIALIZER ---------------------------------
    def __init__(self):
        self.__signal = False
        self.__mean_islet = False
        self.__time = False
        self.__settings = SAMPLE_SETTINGS

        self.__points = False
        self.__cells = False
        self.__filtered_slow = False
        self.__filtered_fast = False
        self.__distributions = False
        self.__binarized_slow = False
        self.__binarized_fast = False
        self.__activity = False
        self.__good_cells = False

# --------------------------------- IMPORTS -----------------------------------
    def import_data(self, signal):
        if not len(signal.shape) == 2:
            raise ValueError("Signal shape not 2D.")
        self.__signal = np.around(signal[:, 1:].transpose(), decimals=3)
        self.__mean_islet = np.mean(self.__signal, 0)  # average over 0 axis
        self.__mean_islet = self.__mean_islet - np.mean(self.__mean_islet)
        sampling = self.__settings["Sampling [Hz]"]
        self.__time = np.arange(len(self.__signal[0]))*(1/sampling)

        self.__points = len(self.__time)
        self.__cells = len(self.__signal)

        self.__good_cells = np.ones(self.__cells, dtype="bool")

    def import_settings(self, settings):
        for key in settings:
            if key in SAMPLE_SETTINGS:
                if isinstance(key, dict):
                    for subkey in key:
                        if subkey in SAMPLE_SETTINGS[key]:
                            self.__settings[subkey] = settings[subkey]
                else:
                    self.__settings[key] = settings[key]
        plateau = self.__settings["Plateau [s]"]
        if plateau[0] != 0 or plateau[1] != 0:
            self.__activity = np.array(
                [plateau for cell in range(self.__cells)]
                )

    def import_good_cells(self, cells):
        if self.__signal is False:
            raise ValueError("No imported data!")
        if len(cells) != self.__cells:
            raise ValueError("Cell number does not match.")
        self.__good_cells = cells

    def reset_computations(self):
        self.__filtered_slow = False
        self.__filtered_fast = False
        self.__distributions = False
        self.__binarized_slow = False
        self.__binarized_fast = False
        self.__activity = False
        self.__good_cells = np.ones(self.__cells, dtype="bool")

# --------------------------------- GETTERS -----------------------------------

    def get_settings(self): return self.__settings
    def get_time(self): return self.__time
    def get_signal(self): return self.__signal
    def get_mean_islet(self): return self.__mean_islet
    def get_points(self): return self.__points
    def get_cells(self): return self.__cells
    def get_filtered_slow(self): return self.__filtered_slow
    def get_filtered_fast(self): return self.__filtered_fast
    def get_distributions(self): return self.__distributions
    def get_binarized_slow(self): return self.__binarized_slow
    def get_binarized_fast(self): return self.__binarized_fast
    def get_activity(self): return self.__activity
    def get_good_cells(self): return self.__good_cells

# ----------------------------- ANALYSIS METHODS ------------------------------
# ---------- Filter + smooth ---------- #

    def filter(self):
        if self.__signal is False:
            raise ValueError("No imported data!")
        slow = self.__settings["Filter"]["Slow [Hz]"]
        fast = self.__settings["Filter"]["Fast [Hz]"]
        self.__filtered_slow = np.zeros((self.__cells, self.__points))
        self.__filtered_fast = np.zeros((self.__cells, self.__points))

        for cell in range(self.__cells):
            self.__filtered_slow[cell] = self.__bandpass(self.__signal[cell],
                                                         (*slow)
                                                         )
            self.__filtered_fast[cell] = self.__bandpass(self.__signal[cell],
                                                         (*fast)
                                                         )
            yield (cell+1)/self.__cells

    def __bandpass(self, data, lowcut, highcut, order=5):
        nyq = 0.5*self.__settings["Sampling [Hz]"]
        low = lowcut / nyq
        high = highcut / nyq
        sos = butter(
            order, [low, high], analog=False, btype='band', output='sos'
            )
        y = sosfiltfilt(sos, data)
        return y

# ---------- Distributions ---------- #

    def compute_distributions(self):
        if self.__filtered_slow is False:
            raise ValueError("No filtered data.")
        self.__distributions = [dict() for i in range(self.__cells)]

        for cell in range(self.__cells):
            signal = self.__filtered_fast[cell]
            signal /= np.max(np.abs(signal))

            # Define noise from time 0 to start of stimulation
            stimulation = int(self.__settings["Stimulation [frame]"][0])
            noise = signal[:int(stimulation)]
            spikes = signal[int(stimulation):]

            # Remove outliers
            q1 = np.quantile(noise, 0.25)
            q3 = np.quantile(noise, 0.75)
            iqr = q3 - q1
            noise = noise[np.logical_and(noise > q1-1.5*iqr,
                                         noise < q3+1.5*iqr
                                         )]

            # Distribution parameters of noise
            noise_params = (skew(noise), np.mean(noise), np.std(noise))
            spikes_params = (skew(spikes), np.mean(spikes), np.std(spikes))

            self.__distributions[cell]["noise_params"] = noise_params
            self.__distributions[cell]["spikes_params"] = spikes_params
            self.__distributions[cell]["noise_hist"] = np.histogram(noise, 20)
            self.__distributions[cell]["spikes_hist"] = np.histogram(
                spikes, 100
                )
            yield (cell+1)/self.__cells

# ---------- Exclude ---------- #

    def autoexclude(self):
        if self.__distributions is False:
            raise ValueError("No distributions.")
        # Excluding thresholds
        score_threshold = self.__settings["Exclude"]["Score threshold"]

        for cell in range(self.__cells):
            skew = self.__distributions[cell]["spikes_params"][0]
            noise_std = self.__distributions[cell]["noise_params"][2]
            spikes_std = self.__distributions[cell]["spikes_params"][2]
            if skew < score_threshold and spikes_std < STD_RATIO*noise_std:
                self.__good_cells[cell] = False
            yield (cell+1)/self.__cells
        print("{} of {} good cells ({:0.0f}%)".format(
            np.sum(self.__good_cells), self.__cells,
            np.sum(self.__good_cells)/self.__cells*100)
            )

    def exclude(self, i):
        if i not in range(self.__cells):
            raise ValueError("Cell not in range.")
        self.__good_cells[i] = False

    def unexclude(self, i):
        if i not in range(self.__cells):
            raise ValueError("Cell not in range.")
        self.__good_cells[i] = True

# ---------- Binarize ---------- #

    def __search_sequence(self, arr, seq):
        # Store sizes of input array and sequence
        seq = np.array(seq)
        Na, Nseq = arr.size, seq.size

        # Range of sequence
        r_seq = np.arange(Nseq)

        # Create a 2D array of sliding indices across the entire length of
        # input array. Match up with the input sequence & get the matching
        # starting indices.
        M = (arr[np.arange(Na-Nseq+1)[:, None] + r_seq] == seq).all(1)

        # Get the range of those indices as final output
        if M.any() > 0:
            return np.where(M)[0]
        else:
            return np.array([], dtype="int")  # No match found

    def binarize_fast(self):
        if self.__distributions is False or self.__filtered_fast is False:
            raise ValueError("No distribution or filtered data.")

        spikes_th = self.__settings["Exclude"]["Spikes threshold"]
        self.__binarized_fast = np.zeros((self.__cells, self.__points), int)
        for cell in range(self.__cells):
            threshold = 3*self.__distributions[cell]["noise_params"][2]
            self.__binarized_fast[cell] = np.where(
                self.__filtered_fast[cell] > threshold, 1, 0
                )
            if np.sum(self.__binarized_fast[cell]) < spikes_th*self.__points:
                self.__good_cells[cell] = False
            yield (cell+1)/self.__cells

    def binarize_slow(self):
        if self.__filtered_slow is False:
            raise ValueError("No filtered data.")
        self.__binarized_slow = np.zeros((self.__cells, self.__points), int)
        for cell in range(self.__cells):
            signal = self.__filtered_slow[cell]
            heavisided_gradient = np.heaviside(np.gradient(signal), 0)
            minima = self.__search_sequence(heavisided_gradient, [0, 1])
            maxima = self.__search_sequence(heavisided_gradient, [1, 0])
            extremes = np.sort(np.concatenate((minima, maxima)))

            reverse_mode = False if minima[0] < maxima[0] else True

            self.__binarized_slow[cell, 0:extremes[0]] = 0
            for i in range(len(extremes)-1):
                e1, e2 = extremes[i], extremes[i+1]
                if i % 2 == int(reverse_mode):
                    self.__binarized_slow[cell, e1:e2] = np.floor(
                        np.linspace(1, 7, e2-e1, endpoint=False)
                        )
                else:
                    self.__binarized_slow[cell, e1:e2] = np.floor(
                        np.linspace(7, 13, e2-e1, endpoint=False)
                        )
            self.__binarized_slow[cell, extremes[-1]:] = 0
            yield (cell+1)/self.__cells
        self.__binarized_slow = self.__binarized_slow.astype(int)

    def autolimit(self):
        if self.__binarized_fast is False:
            raise ValueError("No binarized data.")
        print("Computing activity...")
        self.__activity = []
        for cell in range(self.__cells):
            data = self.__binarized_fast[cell]
            cumsum = np.cumsum(data)

            sampling = self.__settings["Sampling [Hz]"]
            stimulation = self.__settings["Stimulation [frame]"][0]
            lower_limit = cumsum[cumsum < 0.1*cumsum[-1]].size
            lower_limit /= sampling  # lower limit in seconds
            upper_limit = (cumsum.size - cumsum[cumsum > 0.9*cumsum[-1]].size)
            upper_limit /= sampling  # upper limit in seconds

            def box(t, a, t_start, t_end):
                return a*(np.heaviside(t-t_start, 0)-np.heaviside(t-t_end, 0))
            res = differential_evolution(
                lambda p: np.sum((box(self.__time, *p) - data)**2),
                [[0, 100],
                 [0, lower_limit+1],
                 [upper_limit-1, self.__time[-1]]]
                )
            self.__activity.append(res.x[1:])

            if self.__activity[cell][0] < stimulation/sampling:
                self.__good_cells[cell] = False
            yield (cell+1)/self.__cells
        self.__activity = np.array(self.__activity)

    def is_analyzed(self):
        if self.__filtered_slow is False or self.__filtered_fast is False:
            return False
        elif self.__distributions is False:
            return False
        elif self.__binarized_slow is False or self.__binarized_fast is False:
            return False
        elif self.__good_cells is False:
            return False
        elif self.__activity is False:
            return False
        else:
            return True

# ----------------------------- PLOTTING METHODS ------------------------------

    def plot(self, ax, cell, plots=("raw",),
             protocol=True, stimulation=True, noise=False
             ):
        if "mean" in plots:
            signal = self.__mean_islet
            signal = signal/np.max(signal)
            ax.plot(self.__time, signal, "k", alpha=0.25, lw=0.1)
        if "raw" in plots:
            signal = self.__signal[cell]
            signal = signal - np.mean(signal)
            signal = signal/np.max(signal)
            ax.plot(self.__time, signal, "k", alpha=0.5, lw=0.1)
        if "slow" in plots:
            filtered_slow = self.__filtered_slow[cell]
            filtered_slow = filtered_slow/np.max(filtered_slow)
            ax.plot(self.__time, filtered_slow, color="C0", lw=2)
        if "fast" in plots:
            filtered_fast = self.__filtered_fast[cell]
            filtered_fast = filtered_fast/np.max(filtered_fast)
            ax.plot(self.__time, filtered_fast, color="C3", lw=0.2)
        if "bin_slow" in plots:
            binarized_slow = self.__binarized_slow[cell]
            ax2 = ax.twinx()
            ax2.plot(self.__time, binarized_slow, color="k", lw=1)
            ax2.set_ylabel("Phase")
        if "bin_fast" in plots:
            filtered_fast = self.__filtered_fast[cell]
            binarized_fast = self.__binarized_fast[cell]/np.max(filtered_fast)
            binarized_fast *= 3*self.__distributions[cell]["noise_params"][2]
            ax.plot(self.__time, binarized_fast, color="k", lw=1)
            ax2 = ax.twinx()
            ax2.set_ylabel("Action potentials")

        if protocol:
            self.__plot_protocol(ax)

        if stimulation:
            self.__plot_stimulation(ax, cell)

        if noise:
            self.__plot_noise(ax, cell)

        ax.set_xlim(0, self.__time[-1])
        ax.set_ylim(None, 1.1)
        ax.set_xlabel("Time [s]")
        ax.set_ylabel("Amplitude")

    def __plot_protocol(self, ax):
        sampling = self.__settings["Sampling [Hz]"]
        glucose = self.__settings["Glucose [mM]"]
        TA, TAE = self.__settings["Stimulation [frame]"]
        TA, TAE = TA/sampling, TAE/sampling
        if TA == 0 or TAE == 0:
            return
        color = "C0" if glucose == 8 else "C3"

        rectangles = {
            '': patches.Rectangle(
                    (0, 1.1), TA, 0.15, color='grey', alpha=0.5,
                    transform=ax.transData, clip_on=False
                    ),
            '{} mM'.format(glucose): patches.Rectangle(
                                    (TA, 1.1), TAE-TA, 0.3,
                                    color=color, alpha=0.8,
                                    transform=ax.transData, clip_on=False
                                    ),
            '6 mM': patches.Rectangle((TAE, 1.1), self.__time[-1]-TAE,
                                      0.15, color='grey', alpha=0.5,
                                      transform=ax.transData, clip_on=False
                                      )
            }
        for r in rectangles:
            ax.add_artist(rectangles[r])
            rx, ry = rectangles[r].get_xy()
            cx = rx + rectangles[r].get_width()/2.0
            cy = ry + rectangles[r].get_height()/2.0
            ax.annotate(r, (cx, cy), color='k', fontsize=12,
                        ha='center', va='center', xycoords=ax.transData,
                        annotation_clip=False
                        )

    def __plot_stimulation(self, ax, cell):
        frame_start = self.__settings["Stimulation [frame]"][0]
        frame_end = self.__settings["Stimulation [frame]"][1]
        sampling = self.__settings["Sampling [Hz]"]
        ax.axvline(frame_start/sampling, c="grey")
        ax.axvline(frame_end/sampling, c="grey")

        if self.__good_cells[cell]:
            if self.__activity is not False:
                border = self.__activity[cell]
                ax.axvspan(0, border[0], alpha=0.25, color="grey")
                ax.axvspan(
                    border[1], self.__time[-1], alpha=0.25, color="grey"
                    )
        else:
            ax.axvspan(0, self.__time[-1], alpha=0.5, color=EXCLUDE_COLOR)

    def __plot_noise(self, ax, cell):
        noise = 3*self.__distributions[cell]["noise_params"][2]
        ax.fill_between(self.__time, -noise, noise, color="C3", alpha=0.25,
                        label=r"$3\cdot$STD"
                        )

    def plot_distributions(self, ax1, ax2, i):
        if self.__distributions is False:
            raise ValueError("No distribution data.")

        noise_params = self.__distributions[i]["noise_params"]
        spikes_params = self.__distributions[i]["spikes_params"]
        noise_h, noise_bins = self.__distributions[i]["noise_hist"]
        spikes_h, spikes_bins = self.__distributions[i]["spikes_hist"]

        delta_noise = noise_bins[1] - noise_bins[0]
        ax1.bar(noise_bins[:-1], noise_h, delta_noise, color="grey")
        ax1.axvline(noise_params[1], c="k", label="Mean")
        ax1.axvspan(noise_params[1]-noise_params[2],
                    noise_params[1]+noise_params[2],
                    alpha=0.5, color=EXCLUDE_COLOR,
                    label="STD: {:.2f}".format(noise_params[2])
                    )
        ax1.legend()

        delta_spikes = spikes_bins[1]-spikes_bins[0]
        ax2.bar(spikes_bins[:-1], spikes_h, delta_spikes,
                color="grey", label="Skew: {:.2f}".format(spikes_params[0])
                )
        ax2.axvline(spikes_params[1], c="k", label="Mean")
        ax2.axvline(spikes_params[2], c="k", ls="--",
                    label="STD: {:.2f}".format(spikes_params[2])
                    )
        ax2.legend()

    def plot_events(self, ax):
        if self.__binarized_slow is False or self.__binarized_fast is False:
            raise ValueError("No binarized data!")

        bin_fast = self.__binarized_fast[self.__good_cells]
        raster = [[] for i in range(len(bin_fast))]
        sampling = self.__settings["Sampling [Hz]"]

        for i in range(len(bin_fast)):
            for j in range(len(bin_fast[0])):
                yield (i*len(bin_fast[0])+j+1)/len(bin_fast)/len(bin_fast[0])
                if bin_fast[i, j] == 1:
                    raster[i].append(j/sampling)

        ax.eventplot(raster, linewidths=0.1)



import numpy as np
import matplotlib.pyplot as plt

from .networks import Networks


class Analysis(object):
    """docstring for Analysis."""

# ------------------------------- INITIALIZER --------------------------------
    def __init__(self):
        self.__settings = None
        self.__points = None
        self.__cells = None
        self.__positions = None
        self.__filtered_slow = None
        self.__filtered_fast = None
        self.__binarized_slow = None
        self.__binarized_fast = None
        self.__activity = None
        self.__act_sig = None

        self.__networks = False

    def import_data(self, data, positions):
        assert data.is_analyzed()

        good_cells = data.get_good_cells()

        self.__settings = data.get_settings()
        self.__sampling = self.__settings["Sampling [Hz]"]
        self.__points = data.get_points()
        distance = self.__settings["Distance [um]"]
        self.__positions = positions[good_cells]*distance
        self.__cells = np.sum(good_cells)

        self.__filtered_slow = data.get_filtered_slow()[good_cells]
        self.__filtered_fast = data.get_filtered_fast()[good_cells]

        self.__binarized_slow = data.get_binarized_slow()[good_cells]
        self.__binarized_fast = data.get_binarized_fast()[good_cells]

        self.__activity = np.array(data.get_activity())[good_cells]

    def build_networks(self):
        print("Building networks...")
        # Construct networks and build networks from data
        self.__networks = Networks(self.__cells,
                                   self.__filtered_slow,
                                   self.__filtered_fast
                                   )
        self.__networks.build_networks()

# ---------------------------- ANALYSIS FUNCTIONS ----------------------------
    def __search_sequence(self, arr, seq):
        # Store sizes of input array and sequence
        seq = np.array(seq)
        Na, Nseq = arr.size, seq.size

        # Range of sequence
        r_seq = np.arange(Nseq)

        # Create a 2D array of sliding indices across the entire length of
        # input array. Match up with the input sequence & get the matching
        # starting indices.
        M = (arr[np.arange(Na-Nseq+1)[:, None] + r_seq] == seq).all(1)

        # Get the range of those indices as final output
        if M.any() > 0:
            return np.where(M is True)[0]
        else:
            return np.array([], dtype="int")  # No match found

    def __distances_matrix(self):
        A_dst = np.zeros((self.__cells, self.__cells))
        for cell1 in range(self.__cells):
            for cell2 in range(cell1):
                x1, y1 = self.__positions[cell1, 0], self.__positions[cell1, 1]
                x2, y2 = self.__positions[cell2, 0], self.__positions[cell2, 1]
                distance = np.sqrt((x1-x2)**2 + (y1-y2)**2)
                A_dst[cell1, cell2] = distance
                A_dst[cell2, cell1] = distance
        return A_dst

# ------------------------------ GETTER METHODS ------------------------------

    def get_positions(self): return self.__positions
    def get_filtered_slow(self): return self.__filtered_slow
    def get_filtered_fast(self): return self.__filtered_fast
    def get_act_sig(self): return self.__act_sig
    def get_networks(self): return self.__networks

    def get_parameters(self):
        par_cell = [dict() for c in range(self.__cells)]
        par_network = False
        for cell in range(self.__cells):
            par_cell[cell]["AD"] = self.activity(cell)[0]
            par_cell[cell]["AT"] = self.activity(cell)[1]
            par_cell[cell]["OD"] = self.activity(cell)[2]
            par_cell[cell]["Fs"] = self.frequency(cell)[0]
            par_cell[cell]["Ff"] = self.frequency(cell)[1]
            par_cell[cell]["ISI"] = self.interspike(cell)[0]
            par_cell[cell]["ISIV"] = self.interspike(cell)[1]
            par_cell[cell]["TP"] = self.time(cell)["plateau_start"]
            par_cell[cell]["TS"] = self.time(cell)["spike_start"]
            par_cell[cell]["TI"] = self.time(cell)["plateau_end"]
            par_cell[cell]["AMP"] = self.amplitudes()

        if self.__networks is not False:
            par_network = dict()
            par_network["Rs"] = self.average_correlation()[0]
            par_network["Rf"] = self.average_correlation()[1]
            par_network["Ds"] = self.connection_distances()[0]
            par_network["Df"] = self.connection_distances()[1]
            par_network["Qs"] = self.modularity()[0]
            par_network["Qf"] = self.modularity()[1]
            par_network["GEs"] = self.global_efficiency()[0]
            par_network["GEf"] = self.global_efficiency()[1]
            par_network["MCCs"] = self.max_connected_component()[0]
            par_network["MCCf"] = self.max_connected_component()[1]

            for cell in range(self.__cells):
                par_cell[cell]["NDs"] = self.node_degree(cell)[0]
                par_cell[cell]["NDf"] = self.node_degree(cell)[1]
                par_cell[cell]["Cs"] = self.clustering(cell)[0]
                par_cell[cell]["Cf"] = self.clustering(cell)[1]
                par_cell[cell]["NNDs"] = self.nearest_neighbour_degree(cell)[0]
                par_cell[cell]["NNDf"] = self.nearest_neighbour_degree(cell)[1]

# ----------------------- INDIVIDUAL PARAMETER METHODS -----------------------

    def average_correlation(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.average_correlation()

    def connection_distances(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        A_dst = self.__distances_matrix()
        A_slow = self.__networks.get_A_slow()
        A_fast = self.__networks.get_A_fast()

        A_dst_slow = np.multiply(A_dst, A_slow)
        A_dst_fast = np.multiply(A_dst, A_fast)

        slow_distances, fast_distances = [], []
        for c1 in range(self.__cells):
            for c2 in range(c1):
                ds = A_dst_slow[c1, c2]
                df = A_dst_fast[c1, c2]
                if ds > 0:
                    slow_distances.append(ds)
                if df > 0:
                    fast_distances.append(df)

        return (np.array(slow_distances), np.array(fast_distances))

    def modularity(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.modularity()

    def global_efficiency(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.global_efficiency()

    def max_connected_component(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.max_connected_component()

    def amplitudes(self):
        amplitudes = []
        for cell in range(self.__cells):
            heavisided_gradient = np.heaviside(
                np.gradient(self.__filtered_slow[cell]), 0
                )
            minima = self.__search_sequence(heavisided_gradient, [0, 1])
            maxima = self.__search_sequence(heavisided_gradient, [1, 0])

            if maxima[0] < minima[0]:
                maxima = np.delete(maxima, 0)
            if maxima[-1] < minima[-1]:
                minima = np.delete(minima, 0)

            for i, j in zip(minima, maxima):
                amplitudes.append(
                    self.__filtered_slow[cell][j]-self.__filtered_slow[cell][i]
                    )
        return amplitudes

    def activity(self, cell):
        start = int(self.__activity[cell][0]*self.__sampling)
        stop = int(self.__activity[cell][1]*self.__sampling)
        bin = self.__binarized_fast[cell][start:stop]
        sum = np.sum(bin)
        length = bin.size
        Nf = self.frequency(cell)[1]*length/self.__sampling
        if sum == 0:
            return (length, np.nan, np.nan)
        return (length/self.__sampling, sum/length, (sum/self.__sampling)/Nf)

    def frequency(self, cell):
        start = int(self.__activity[cell][0]*self.__sampling)
        stop = int(self.__activity[cell][1]*self.__sampling)
        bin_slow = self.__binarized_slow[cell][start:stop]
        bin_fast = self.__binarized_fast[cell][start:stop]

        slow_peaks = self.__search_sequence(bin_slow, [11, 12])
        if slow_peaks.size < 2:
            frequency_slow = np.nan
        else:
            slow_interval = slow_peaks[-1]-slow_peaks[0]
            frequency_slow = (slow_peaks.size-1)/slow_interval*self.__sampling

        fast_peaks = self.__search_sequence(bin_fast, [0, 1])
        if fast_peaks.size < 2:
            frequency_fast = np.nan
        else:
            fast_interval = fast_peaks[-1]-fast_peaks[0]
            frequency_fast = (fast_peaks.size-1)/fast_interval*self.__sampling

        return (frequency_slow, frequency_fast)

    def interspike(self, cell):
        start = int(self.__activity[cell][0]*self.__sampling)
        stop = int(self.__activity[cell][1]*self.__sampling)
        bin_fast = self.__binarized_fast[cell][start:stop]

        IS_start = self.__search_sequence(bin_fast, [1, 0])
        IS_end = self.__search_sequence(bin_fast, [0, 1])

        if IS_start.size == 0 or IS_end.size == 0:
            return (np.nan, np.nan)
        # First IS_start must be before first interspike_end
        if IS_end[-1] < IS_start[-1]:
            IS_start = IS_start[:-1]
        if IS_start.size == 0:
            return (np.nan, np.nan)
        # Last IS_start must be before last interspike_end
        if IS_end[0] < IS_start[0]:
            IS_end = IS_end[1:]

        assert IS_start.size == IS_end.size

        IS_lengths = [IS_end[i]-IS_start[i] for i in range(IS_start.size)]
        mean_IS_interval = np.mean(IS_lengths)
        IS_variation = np.std(IS_lengths)/mean_IS_interval

        return (mean_IS_interval, IS_variation)

    def time(self, cell):
        bin_fast = self.__binarized_fast[cell]
        time = {}
        stim_start = self.__settings["Stimulation [frame]"][0]
        stim_end = self.__settings["Stimulation [frame]"][1]

        time["plateau_start"] = self.__activity[cell][0] - stim_start
        time["plateau_end"] = self.__activity[cell][1] - stim_end

        fast_peaks = self.__search_sequence(bin_fast[stim_start:], [0, 1])
        if len(fast_peaks) < 3:
            time["spike_start"] = np.nan
        else:
            time["spike_start"] = (np.mean(fast_peaks[:3]))/self.__sampling

        return time

    def node_degree(self, cell):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.node_degree(cell)

    def clustering(self, cell):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.clustering(cell)

    def nearest_neighbour_degree(self, cell):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.nearest_neighbour_degree(cell)

# ----------------------------- ANALYSIS METHODS ------------------------------
# ----------------------------- Spikes vs phases ------------------------------

    def spikes_vs_phase(self, mode="normal"):
        phases = np.arange((np.pi/3 - np.pi/6)/2, 2*np.pi, np.pi/6)
        spikes = np.zeros((self.__cells, 12))

        # Iterate through cells
        for cell in range(self.__cells):
            start = int(self.__activity[cell][0]*self.__sampling)
            stop = int(self.__activity[cell][1]*self.__sampling)

            bin_slow = self.__binarized_slow[cell][start:stop]
            bin_fast = self.__binarized_fast[cell][start:stop]

            # Iterate through phases (1–12)
            for phase in range(1, 13):
                # Bool array with True at slow phase:
                slow_isolated = bin_slow == phase

                # Bool array with True at fast spike:
                spike_indices = self.__search_sequence(bin_fast, [0, 1]) + 1
                fast_unitized = np.zeros(len(bin_fast))
                fast_unitized[spike_indices] = 1

                # Bool array with True at fast spike AND slow phase
                fast_isolated_unitized = np.logical_and(
                    slow_isolated, fast_unitized
                    )

                # Append result
                spikes[cell, phase-1] = np.sum(fast_isolated_unitized)

        if mode == "normal":
            result = np.sum(spikes, axis=0)
        elif mode == "separate":
            result = spikes
        return (phases, result)

# ------------------------- Correlation vs distance ---------------------------

    def correlation_vs_distance(self):
        A_dst = self.__distances_matrix()
        distances = []
        correlations_slow = []
        correlations_fast = []
        for cell1 in range(self.__cells):
            for cell2 in range(cell1):
                distances.append(A_dst[cell1, cell2])
                corr_slow = np.corrcoef(self.__filtered_slow[cell1],
                                        self.__filtered_slow[cell2])[0, 1]
                corr_fast = np.corrcoef(self.__filtered_fast[cell1],
                                        self.__filtered_fast[cell2])[0, 1]
                correlations_slow.append(corr_slow)
                correlations_fast.append(corr_fast)
        return (distances, correlations_slow, correlations_fast)

# -------------------------- WAVE DETECTION METHODS ---------------------------
    def wave_detection(self, time_th=0.5):
        print("Detecting waves")
        event_num = []

        bin_sig = self.__binarized_fast
        act_sig = np.zeros_like(bin_sig, int)
        frame_th = int(time_th*self.__sampling)
        R = self.__distances_matrix()
        R_th = np.average(R) - np.std(R)

        neighbours = []
        for i in range(self.__cells):
            neighbours.append(np.where((R[i, :] < R_th) & (R[i, :] != 0))[0])

        nonzero = {}
        # Poisce vse frejme, kjer je kakšna celica aktivna
        nonzero_frames = np.where(bin_sig.T == 1)[0]
        # zanka po frejmih z aktivnostjo
        for frame in nonzero_frames:
            # v frejmu z aktivnostjo poisce vse celice, ki so dejansko aktivne
            nonzero[frame] = list(np.where(bin_sig.T[frame, :] == 1)[0])

        counter = 0
        # zanka po frame-ih z aktivnostjo
        for frame in nonzero:
            # Obdela prvi aktivni frame v binsig matriki
            if counter == 0:
                k = 1
                # Zanka po aktivnih celicah v frejmu
                for cell in nonzero[frame]:
                    act_sig[cell, frame] = k
                    k += 1

                for nn in nonzero[frame]:
                    current = set(nonzero[frame])
                    neighbours_nn = set(neighbours[nn])
                    for nnn in list(neighbours_nn.intersection(current)):
                        act_sig[nn, frame] = min(act_sig[nn, frame],
                                                 act_sig[nnn, frame]
                                                 )
                        act_sig[nnn, frame] = act_sig[nn, frame]

                un_num = np.unique(act_sig[:, frame])
                event_num = list(set(event_num).union(set(un_num)))

                max_event_num = max(event_num)
                counter += 1

            if counter != 0:
                k = max_event_num + 1
                # Zanka po aktivnih celicah
                for cell in nonzero[frame]:
                    # Če je celica v tem frameu aktivna, v prejsnjem pa ne, ji
                    # dodeli novo številko dogodka
                    if bin_sig[cell, frame-1] == 0:
                        act_sig[cell, frame] = k
                        k += 1
                    # Če je celica bila že v prejsnjem frejmu aktivna, ji
                    # prepiše indeks dogodka
                    else:
                        act_sig[cell, frame] = act_sig[cell, frame-1]

                for nn in nonzero[frame]:
                    current = set(nonzero[frame])
                    neighbours_nn = set(neighbours[nn])
                    for nnn in list(neighbours_nn.intersection(current)):
                        if act_sig[nn, frame] != 0 and \
                                act_sig[nnn, frame] != 0 and \
                                act_sig[nn, frame-1] != 0 and \
                                np.sum(bin_sig[nn, frame-frame_th:frame+1]) \
                                <= frame_th and \
                                act_sig[nnn, frame-1] == 0 and \
                                nn != nnn:
                            act_sig[nnn, frame] = act_sig[nn, frame]
                        if act_sig[nn, frame] != 0 and \
                                act_sig[nnn, frame] != 0 and \
                                act_sig[nn, frame-1] == 0 and \
                                act_sig[nnn, frame-1] != 0 and \
                                np.sum(bin_sig[nnn, frame-frame_th:frame+1]) \
                                <= frame_th and \
                                nn != nnn:
                            act_sig[nn, frame] = act_sig[nnn, frame]
                        if act_sig[nn, frame] != 0 and \
                                act_sig[nnn, frame] != 0 and \
                                act_sig[nn, frame-1] == 0 and \
                                act_sig[nnn, frame-1] == 0 and \
                                nn != nnn:
                            act_sig[nn, frame] = min(act_sig[nn, frame],
                                                     act_sig[nnn, frame]
                                                     )
                            act_sig[nnn, frame] = act_sig[nn, frame]

                un_num = np.unique(act_sig[:, frame])
                event_num = list(set(event_num).union(set(un_num)))

                max_event_num = max(event_num)
                counter += 1
        self.__act_sig = act_sig

    def wave_characterization(self, big_th=0.45, small_th=0.1, time_th=0.5):
        if self.__act_sig is None:
            self.wave_detection(time_th)
        print("Characterizing waves")
        # vse stevilke dogodkov razen nicle - 0=neaktivne celice
        events = np.unique(self.__act_sig[self.__act_sig != 0])
        # print(events)
        # print(events.size, np.min(events), np.max(events))

        big_events = []
        all_events = []

        for e in events:
            e = int(e)
            cells, frames = np.where(self.__act_sig == e)
            active_cell_number = np.unique(cells).size

            start_time, end_time = np.min(frames), np.max(frames)

            characteristics = {
                "event number": e,
                "start time": start_time,
                "end time": end_time,
                "active cell number": active_cell_number,
                "rel active cell number": active_cell_number/self.__cells
            }
            if active_cell_number > int(big_th*self.__cells):
                big_events.append(characteristics)
            if active_cell_number > int(small_th*self.__cells):
                all_events.append(characteristics)

        return (big_events, all_events)

# ------------------------------ DRAWING METHODS ------------------------------

    def draw_networks(self, ax1, ax2, colors):
        return self.__networks.draw_networks(
            self.__positions, ax1, ax2, colors
            )

    def plot_events(self, events, all_events):
        for e in (events, all_events):
            rast_plot = []
            zacetki = []
            k = 0
            kk = 0
            for c in e:
                zacetki.append([])
                event_num = int(c["event number"])
                start_time = int(c["start time"])
                end_time = int(c["end time"])
                active_cell_number = c["active cell number"]

                step = 0
                used = []
                init_cells = 0
                for i in range(self.__cells):
                    for j in range(start_time, end_time+1, 1):
                        if self.__act_sig[i, j] == event_num and i not in used:
                            rast_plot.append([])
                            rast_plot[k].append(
                                (start_time+step)/self.__sampling
                                )
                            rast_plot[k].append(i)
                            rast_plot[k].append(event_num)
                            rast_plot[k].append(active_cell_number)
                            used.append(i)
                            k += 1
                    init_cells += 1
                    step += 1

                zacetki[kk].append(start_time/self.__sampling)
                zacetki[kk].append(-5)
                zacetki[kk].append(event_num)
                kk += 1

            fzacetki = np.array(zacetki, float)
            frast_plot = np.array(rast_plot, float)

            fig = plt.figure(figsize=(8, 4))
            ax = fig.add_subplot(111)
            ax.scatter(frast_plot[:, 0], frast_plot[:, 1],
                       s=0.5, c=frast_plot[:, 2], marker='o'
                       )
            ax.scatter(fzacetki[:, 0], fzacetki[:, 1], s=10.0, marker='+')
            ax.set_xlabel('Time (s)')
            ax.set_ylabel('Cell $i$')
            plt.show()



from setuptools import setup, find_packages

setup(
      name = "langerhans",
      version = "1.6.0",
      description = "Analyzer for the calcium signals of Islets of Langerhans.",
      author = "Jan Zmazek",
      url = "https://github.com/janzmazek/langerhans",
      license = "MIT License",
      packages = find_packages(exclude=['*test']),
      install_requires = ['numpy', 'scipy', 'matplotlib', 'networkx', 'pyyaml', 'python-louvain']
)
